import pytest

from rail.projects import name_utils


def test_name_utils() -> None:
    assert name_utils.get_required_interpolants("xx_{alice}_{bob}") == [
        "{alice}",
        "{bob}",
    ]
    assert (
        name_utils.format_template("xx_{alice}_{bob}", alice="x", bob="x") == "xx_x_x"
    )

    test_dict = dict(
        a="a_{alice}",
        # b=['b1_{alice}', 'b2_{bob}'],
        c=dict(c1="c1_{alice}", c2="c2_{alice}"),
        # c2=dict(c2_1='c1_{alice}', c2_2=['c2_{alice}', 'c2_{bob}']),
    )

    name_utils.resolve_dict(test_dict, dict(alice="x", bob="y"))

    assert not name_utils.resolve_dict(None, {})
    with pytest.raises(ValueError):
        name_utils.resolve_dict(
            dict(
                a=(
                    "s",
                    "d",
                )
            ),
            dict(alice="x", bob="y"),
        )


def test_name_factory() -> None:
    null_factory = name_utils.NameFactory()

    with pytest.raises(KeyError):
        null_factory.get_template("nope", "nope")

    simple_factory = name_utils.NameFactory(
        templates=dict(
            alice="{bob}_{bob}",
        ),
    )

    assert simple_factory.resolve_template("CommonPaths", "root") == "."

    simple_factory.interpolants = dict(root="xx")
    assert simple_factory.interpolants["root"] == "xx"
    del simple_factory.interpolants
    assert simple_factory.interpolants.get("root") is None



from __future__ import annotations

import itertools
import os
from typing import Any, Type, cast

import yaml
from ceci.config import StageParameter
from rail.core.configurable import Configurable
from rail.core.model import Model

from . import execution, library, name_utils
from .algorithm_factory import RailAlgorithmFactory
from .catalog_factory import RailCatalogFactory
from .catalog_template import RailProjectCatalogTemplate
from .file_template import RailProjectFileTemplate
from .pipeline_factory import RailPipelineFactory
from .pipeline_holder import RailPipelineTemplate
from .project_file_factory import RailProjectFileFactory
from .reducer import RailReducer
from .selection_factory import RailSelection, RailSelectionFactory
from .subsample_factory import RailSubsample, RailSubsampleFactory


class RailFlavor(Configurable):
    """Description of a single analysis variation

    This includes

    a name for the variant, used to construct filenames

    a 'catalog_tag', which identifies format of the data being used, and sets
    the expected names of columns accordingly

    a list of 'pipelines' that can be run in this variant

    a list of 'file_aliases' that can be used to specify the
    input files used in this variant

    a dict of 'pipeline_overrides' that modify the behavior of the various pipelines
    """

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Flavor name"),
        catalog_tag=StageParameter(
            str, None, fmt="%s", required=False, msg="tag for catalog being used"
        ),
        pipelines=StageParameter(list, ["all"], fmt="%s", msg="pipelines being used"),
        file_aliases=StageParameter(dict, {}, fmt="%s", msg="file aliases used"),
        pipeline_overrides=StageParameter(dict, {}, fmt="%s", msg="file aliases used"),
    )

    def __init__(self, **kwargs: Any) -> None:
        """C'tor

        Parameters
        ----------
        **kwargs:
            Configuration parameters for this RailFlavor, must match
            class.config_options data members
        """
        Configurable.__init__(self, **kwargs)


class RailProject(Configurable):  # pylint: disable=too-many-public-methods
    """Main analysis driver class, this collects all the elements needed to
    run a collection of studies using RAIL.

    The key concepts are:

    1. analysis 'Flavors', which are versions of
    similar analyses with slightly different parameter settings and/or
    input files.

    2. ceci 'Pipelines', which run blocks of analysis code

    A RailProject basically specifies which Pipelines to run under which
    flavors, and keeps track of the outputs.

    RailProject.functionality_help() for more on class functionality

    RailProject.configuration_help() for more on class configuration
    """

    config_options: dict[str, StageParameter] = dict(
        Name=StageParameter(str, None, fmt="%s", required=True, msg="Project name"),
        Includes=StageParameter(list, [], fmt="%s", msg="Files to include"),
        Baseline=StageParameter(
            dict, None, fmt="%s", required=True, msg="Baseline analysis configuration"
        ),
        Flavors=StageParameter(list, [], fmt="%s", msg="Analysis variants"),
        PathTemplates=StageParameter(dict, {}, fmt="%s", msg="File path templates"),
        CommonPaths=StageParameter(
            dict, {}, fmt="%s", required=True, msg="Paths to shared directories"
        ),
        IterationVars=StageParameter(
            dict, {}, fmt="%s", msg="Iteration variables to use"
        ),
        Selections=StageParameter(
            list, ["all"], fmt="%s", msg="Data selections to use"
        ),
        Subsamples=StageParameter(
            list, ["all"], fmt="%s", msg="Subsample defintions to use"
        ),
        Catalogs=StageParameter(
            list, ["all"], fmt="%s", msg="Catalog templates to use"
        ),
        Files=StageParameter(list, ["all"], fmt="%s", msg="Catalog templates to use"),
        Pipelines=StageParameter(
            list, ["all"], fmt="%s", msg="Catalog templates to use"
        ),
        Reducers=StageParameter(list, ["all"], fmt="%s", msg="Data reducers to use"),
        Subsamplers=StageParameter(
            list, ["all"], fmt="%s", msg="Data subsamplers to use"
        ),
        PZAlgorithms=StageParameter(
            list, ["all"], fmt="%s", msg="p(z) algorithms to use"
        ),
        SpecSelections=StageParameter(
            list, ["all"], fmt="%s", msg="Spectroscopic selections to use"
        ),
        Classifiers=StageParameter(
            list, ["all"], fmt="%s", msg="Tomographic classifiers to use"
        ),
        Summarizers=StageParameter(
            list, ["all"], fmt="%s", msg="n(z) summarizers to use"
        ),
        ErrorModels=StageParameter(
            list, ["all"], fmt="%s", msg="Photometric ErrorModels to use"
        ),
    )

    yaml_tag: str = "Project"

    projects: dict[str, RailProject] = {}

    @classmethod
    def functionality_help(cls) -> None:
        """
        The main functions that the use will use using are:

        load_config:
        ------------
        Read a yaml file and create a RailProject

        reduce_data:
        ------------
        Make a reduced catalog from an input catalog by applying a selction
        and trimming unwanted colums.  This is run before the analysis pipelines.

        subsample_data:
        ---------------
        Subsample data from a catalog to make a testing or training file.
        This is run after catalog level pipelines, but before pipeliens run
        on indvidudal training/ testing samples

        split_data:
        ---------------
        Split data from an input file to make testing and training file.
        This is run after catalog level pipelines, and subsample_data, but
        before pipeliens run on indvidudal training/ testing samples.
        This is an alternative to using subsample_data to make these files.

        build_pipelines:
        ----------------
        Build ceci pipeline yaml files

        run_pipeline_single:
        --------------------
        Run a pipeline on a single file

        run_pipeline_catalog:
        ---------------------
        Run a pipeline on a catalog of files
        """
        print(cls.functionality_help.__doc__)

    @classmethod
    def configuration_help(cls) -> None:
        """
        Configuring a RailProject

        Most of these element come from the shared library of elements,
        which is accesible from rail.projects.library

        ==========================
        Shared configuration files
        ==========================

        Includes: list[str]
        -------------------
        List of shared configuration files to load

        ========================
        Project analysis flavors
        ========================

        See :py:class:`rail.projects.project.RailFlavor` for the parameters
        needed to define an analysis 'Flavor'.

        Baseline: dict[str, Any]
        ------------------------
        Baseline configuration for this project.
        This is included in all the other analysis flavors

        Flavors: list[dict[str, Any]]
        -----------------------------
        List of all the analysis flavors that have been defined in this project

        ====================
        Bookkeeping elements
        ====================

        These are used to define the file paths for the project.

        PathTemplates: dict[str, str]
        -----------------------------
        Overrides for templates used to construct file paths

        The defaults are given in :py:mod:`rail.projects.name_utils`

        .. code-block:: python

            PathTemplates = dict(
                pipeline_path="{pipelines_dir}/{pipeline}_{flavor}.yaml",
                ceci_output_dir="{project_dir}/data/{selection}_{flavor}",
                ceci_file_path="{tag}_{stage}.{suffix}",
            )

        CommonPaths: dict[str, str]
        ---------------------------
        Defintions of common paths used to construct file paths

        The defaults are given in :py:mod:`rail.projects.name_utils`

        .. code-block:: python

            CommonPaths = dict(
                root=".",          # needs to be overridden
                scratch_root=".",  # needs to be overridden
                project="",        # needs to be overridden
                project_dir="{root}/projects/{project}",
                project_scratch_dir="{scratch_root}/projects/{project}",
                catalogs_dir="{root}/catalogs",
                pipelines_dir="{project_dir}/pipelines",
            )

        IterationVars: dict[str, list[str]]
        -----------------------------------
        Iteration variables to construct the catalogs


        ===============
        Shared elements
        ===============
        Things that are pulled from the library, each of these is just a list
        of the names of things that are defined in the library that
        can be used in this project.  The default is to use all the
        items defined in the library.

        Catalogs: list[str] These are actually CatalogTemplates

        Files: list[str] These are actually FileTemplates

        Pipelines: list[str] These are actually PipelineTemplates

        Reducers: list[str] These reduce the input data catalog

        Subsamplers: list[str] These subsample catalogs to get individual files

        Selections: list[str] These are the selection parameters

        Subsamples: list[str] These are the subsample parameters

        PZAlgorithms: list[str]

        SpecSelections: list[str]

        Classifiers: list[str]

        Summarizers: list[str]

        ErrorModels: list[str]
        """
        print(cls.configuration_help.__doc__)

    def __init__(self, **kwargs: Any) -> None:
        """C'tor

        Parameters
        ----------
        **kwargs:
            Configuration parameters for this RailProject, must match
            class.config_options data members
        """
        Configurable.__init__(self, **kwargs)

        for include_ in self.config.Includes:
            library.load_yaml(os.path.expandvars(include_))

        self.name_factory = name_utils.NameFactory(
            config=self.config,
            templates=self.config.PathTemplates,
            interpolants=self.config.CommonPaths,
        )
        self._iteration_vars: dict[str, list[Any]] | None = None
        self._catalog_templates: dict[str, RailProjectCatalogTemplate] | None = None
        self._file_templates: dict[str, RailProjectFileTemplate] | None = None
        self._pipeline_templates: dict[str, RailPipelineTemplate] | None = None
        self._algorithms: dict[str, dict[str, dict[str, str]]] = {}
        self._selections: dict[str, RailSelection] | None = None
        self._subsamples: dict[str, RailSubsample] | None = None
        self._flavors: dict[str, RailFlavor] | None = None

    def __repr__(self) -> str:
        return f"{self.config.Name}"

    def clear_cache(self) -> None:
        """Reset all the cached configurable items"""
        self._catalog_templates = None
        self._file_templates = None
        self._pipeline_templates = None
        self._algorithms = {}
        self._selections = None
        self._subsamples = None
        self._flavors = None

    @staticmethod
    def generate_kwargs_iterable(**iteration_dict: Any) -> list[dict]:
        """Generate a list of kwargs dicts from a dict of lists"""
        iteration_vars = list(iteration_dict.keys())
        iterations = itertools.product(
            *[iteration_dict.get(key, []) for key in iteration_vars]
        )
        iteration_kwarg_list = []
        for iteration_args in iterations:
            iteration_kwargs = {
                iteration_vars[i]: iteration_args[i] for i in range(len(iteration_vars))
            }
            iteration_kwarg_list.append(iteration_kwargs)
        return iteration_kwarg_list

    @staticmethod
    def generate_ceci_command(
        pipeline_path: str,
        config: str | None,
        inputs: dict,
        output_dir: str = ".",
        log_dir: str = ".",
        **kwargs: Any,
    ) -> list[str]:
        """Generate a ceci command to run a pipeline

        Parameters
        ----------
        pipeline_path: str
            Path to the pipline yaml file

        config: str | None
            Path to the pipeline config yaml file

        inputs: dict
            Input to the pipeline

        output_dir: str, default="."
            Pipeline output directory

        log_dir: str, default="."
            Pipeline log directory

        **kwargs:
            These are appended to the command in key=value pairs
        """

        if config is None:
            config = pipeline_path.replace(".yaml", "_config.yml")

        command_line = [
            "ceci",
            f"{pipeline_path}",
            f"config={config}",
            f"output_dir={output_dir}",
            f"log_dir={log_dir}",
        ]

        for key, val in inputs.items():
            command_line.append(f"inputs.{key}={val}")

        for key, val in kwargs.items():
            command_line.append(f"{key}={val}")

        return command_line

    @property
    def name(self) -> str:
        return self.config.Name

    @staticmethod
    def load_config(config_file: str) -> RailProject:
        """Create and return a RailProject from a yaml config file"""
        with open(os.path.expandvars(config_file), "r", encoding="utf-8") as fp:
            config_orig = yaml.safe_load(fp)

        project_config = config_orig.get("Project")
        project = RailProject(**project_config)
        RailProject.projects[project.name] = project
        return project

    def reduce_data(
        self,
        catalog_template: str,
        output_catalog_template: str,
        reducer_class_name: str,
        input_selection: str,
        selection: str,
        dry_run: bool = False,
        **kwargs: Any,
    ) -> list[str]:
        """Reduce some data

        Parameters
        ----------
        catalog_template: str
            Tag for the input catalog

        output_catalog_template: str
            Which label to apply to output dataset

        reducer_class_name: str,
            Name of the class to use for subsampling

        input_selection: str,
            Selection to use for the input

        selection: str,
            Selection to apply

        dry_run: bool
            If true, do not actually run

        **kwargs:
            Used to provide values for additional interpolants.

        Returns
        -------
        list[str]:
            Paths to output files

        """
        sources = self.get_catalog_files(
            catalog_template, selection=input_selection, **kwargs
        )
        sinks = self.get_catalog_files(
            output_catalog_template, selection=selection, **kwargs
        )

        reducer_class = library.get_algorithm_class(
            "Reducer", reducer_class_name, "Reduce"
        )
        assert issubclass(reducer_class, RailReducer)

        reducer_args = library.get_selection(selection)
        reducer = reducer_class(**reducer_args.config.to_dict())

        if not dry_run:  # pragma: no cover
            for source_, sink_ in zip(sources, sinks):
                reducer.run(source_, sink_)

        return sinks

    def subsample_data(
        self,
        catalog_template: str,
        file_template: str,
        subsampler_class_name: str,
        subsample_name: str,
        dry_run: bool = False,
        **kwargs: Any,
    ) -> str:
        """Subsammple some data

        Parameters
        ----------
        catalog_template: str
            Tag for the input catalog

        file_template: str
            Which label to apply to output dataset

        subsampler_class_name: str,
            Name of the class to use for subsampling

        subsample_name: str,
            Name of the subsample to create

        dry_run: bool
            If true, do not actually run

        **kwargs:
            Used to provide values for additional interpolants, e.g., flavor, basename, etc...

        Returns
        -------
        str:
            Path to output file
        """
        hdf5_output = self.get_file(file_template, **kwargs)
        output = hdf5_output.replace(".hdf5", ".parquet")

        subsampler_class = library.get_algorithm_class(
            "Subsampler", subsampler_class_name, "Subsample"
        )
        subsampler_config_keys = cast(
            Type[Configurable], subsampler_class
        ).config_options.keys()
        subsampler_args = library.get_subsample(subsample_name)

        use_pairs = {
            key: val
            for key, val in subsampler_args.config.to_dict().items()
            if key in subsampler_config_keys
        }

        subsampler = subsampler_class(**use_pairs)

        basename_dict: dict[str, str] = subsampler.get_basename_dict()
        sources_dict: dict[str, list[str]] = {}

        for key, val in basename_dict.items():
            kwargs_copy = kwargs.copy()
            kwargs_copy["basename"] = val
            sources_dict[key] = self.get_catalog_files(catalog_template, **kwargs_copy)

        # output_dir = os.path.dirname(output)
        if not dry_run:  # pragma: no cover
            subsampler.run(sources_dict, output)

        return output

    def split_data(
        self,
        file_template: str,
        test_file_template: str,
        train_file_template: str,
        splitter_class_name: str,
        dry_run: bool = False,
        **kwargs: Any,
    ) -> list[str]:
        """Subsammple some data

        Parameters
        ----------
        file_template: str
            Which label to apply to output dataset

        train_file_template: str
            Which label to apply to train output dataset

        test_file_template: str
            Which label to apply to test output dataset

        splitter_class_name: str,
            Name of the class to use for splitting

        dry_run: bool
            If true, do not actually run

        **kwargs:
            Used to provide values for additional interpolants, e.g., flavor, basename, etc...

        Returns
        -------
        str:
            Path to output file
        """
        input_file = self.get_file(file_template, **kwargs).replace(".hdf5", ".parquet")
        output_train_file = self.get_file(train_file_template, **kwargs).replace(
            ".hdf5", ".parquet"
        )
        output_test_file = self.get_file(test_file_template, **kwargs).replace(
            ".hdf5", ".parquet"
        )

        splitter_class = library.get_algorithm_class(
            "Splitter", splitter_class_name, "Split"
        )
        splitter = splitter_class(name=splitter_class_name)

        # output_dir = os.path.dirname(output)
        if not dry_run:  # pragma: no cover
            splitter.run(input_file, output_train_file, output_test_file)

        return [output_train_file, output_test_file]

    def build_pipelines(
        self,
        flavor: str = "baseline",
        *,
        force: bool = False,
    ) -> int:
        """Build ceci pipeline configuraiton files for this project

        Parameters
        ----------
        flavor: str
            Which analysis flavor to draw from

        force: bool
            Force overwriting of existing pipeline files

        Returns
        -------
        int:
            0 if ok, error code otherwise
        """
        flavor_dict = self.get_flavor(flavor)
        pipelines_to_build = flavor_dict["pipelines"]
        all_flavor_overrides = flavor_dict.get("pipeline_overrides", {}).copy()
        do_all = "all" in pipelines_to_build

        ok = 0
        for pipeline_name, pipeline_info in self.get_pipelines().items():
            if not (do_all or pipeline_name in pipelines_to_build):
                print(f"Skipping pipeline {pipeline_name} from flavor {flavor}")
                continue
            output_yaml = self.get_path(
                "pipeline_path", pipeline=pipeline_name, flavor=flavor
            )
            if os.path.exists(output_yaml):  # pragma: no cover
                if force:
                    print(f"Overwriting existing pipeline {output_yaml}")
                else:
                    print(f"Skipping existing pipeline {output_yaml}")
                    continue

            overrides = all_flavor_overrides.get("default", {}).copy()
            pipeline_overrides = all_flavor_overrides.get(pipeline_name, {}).copy()
            overrides.update(**pipeline_overrides)
            pipeline_instance = pipeline_info.make_instance(self, flavor, overrides)
            ok |= pipeline_instance.build(self)

        return ok

    def make_pipeline_single_input_command(
        self,
        pipeline_name: str,
        flavor: str,
        **kwargs: Any,
    ) -> list[str]:
        """Build the command to run pipeline on a single file

        Parameters
        ----------
        pipeline_name: str
            Pipeline in question

        flavor: str
            Flavor to apply

        **kwargs: Any
            Other interpolants, such as selection

        Returns
        -------
        list[str]:
            Tokens in the command line, usable by subprocess.run()
        """
        pipeline_template = self.get_pipeline(pipeline_name)
        pipeline_instance = pipeline_template.make_instance(self, flavor, {})
        return pipeline_instance.make_pipeline_single_input_command(self, **kwargs)

    def make_pipeline_catalog_commands(
        self,
        pipeline_name: str,
        flavor: str,
        **kwargs: Any,
    ) -> list[tuple[list[list[str]], str]]:
        """Build the commands to run pipeline on a catalog

        Parameters
        ----------
        pipeline_name: str
            Pipeline in question

        flavor: str
            Flavor to apply

        **kwargs: Any
            Other interpolants, such as selection

        Returns
        -------
        list[tuple[list[list[str]], str]:
            List of pairs of series of commands and potential location for slurm batch file
        """
        pipeline_template = self.get_pipeline(pipeline_name)
        pipeline_instance = pipeline_template.make_instance(self, flavor, {})
        return pipeline_instance.make_pipeline_catalog_commands(self, **kwargs)

    def run_pipeline_single(
        self,
        pipeline_name: str,
        run_mode: execution.RunMode = execution.RunMode.bash,
        **kwargs: Any,
    ) -> int:
        """Run pipeline on a single file

        Parameters
        ----------
        pipeline_name: str
            Pipeline in question

        run_mode: execution.RunMode
            How to run the pipeline (e.g., in bash, or in slurm)

        **kwargs: Any
            Other interpolants, such as selection

        Returns
        -------
        int:
            0 for success, error code otherwise
        """
        kwcopy = kwargs.copy()
        flavor = kwcopy.pop("flavor")
        sink_dir = self.get_path("ceci_output_dir", flavor=flavor, **kwcopy)
        script_path = os.path.join(sink_dir, f"submit_{pipeline_name}.sh")
        commands = self.make_pipeline_single_input_command(
            pipeline_name, flavor, **kwcopy
        )
        try:
            statuscode = execution.handle_commands(
                run_mode,
                [commands],
                script_path,
                site=kwargs.get("site", None),
            )
        except Exception as msg:  # pragma: no cover
            print(msg)
            statuscode = 1
        return statuscode

    def run_pipeline_catalog(
        self,
        pipeline_name: str,
        run_mode: execution.RunMode = execution.RunMode.bash,
        **kwargs: Any,
    ) -> int:
        """Run pipeline on a catalog

        Parameters
        ----------
        pipeline_name: str
            Pipeline in question

        run_mode: execution.RunMode
            How to run the pipeline (e.g., in bash, or in slurm)

        **kwargs: Any
            Other interpolants, such as selection

        Returns
        -------
        int:
            0 for success, error code otherwise
        """

        kwcopy = kwargs.copy()
        flavor = kwcopy.pop("flavor")
        all_commands = self.make_pipeline_catalog_commands(
            pipeline_name, flavor, **kwcopy
        )

        ok = 0
        for commands, script_path in all_commands:
            try:
                execution.handle_commands(
                    run_mode,
                    commands,
                    script_path,
                    site=kwargs.get("site", None),
                )
            except Exception as msg:  # pragma: no cover
                print(msg)
                ok |= 1

        return ok

    def add_flavor(self, name: str, **kwargs: Any) -> RailFlavor:
        """Add a new flavor to the Project"""
        if self._flavors is None:  # pragma: no cover
            self.get_flavors()
        assert self._flavors is not None
        if name in self._flavors:
            raise KeyError(f"Flavor {name} already in RailProject {self.name}")
        flavor_params = self.config.Baseline.copy()
        flavor_params.update(name=name, **kwargs)
        new_flavor = RailFlavor(**flavor_params)
        self.config["Flavors"].append(new_flavor.config.to_dict())
        self._flavors[new_flavor.config.name] = new_flavor
        return new_flavor

    def write_yaml(self, yaml_file: str) -> None:
        """Write this project to a yaml file"""
        the_dict = self.to_yaml_dict()
        with open(os.path.expandvars(yaml_file), "w", encoding="utf-8") as fout:
            yaml.dump(the_dict, fout)

    def get_path_templates(self) -> dict:
        """Return the dictionary of templates used to construct paths"""
        return self.name_factory.get_path_templates()

    def get_path(self, path_key: str, **kwargs: Any) -> str:
        """Resolve and return a path using the kwargs as interopolants"""
        return self.name_factory.resolve_path_template(path_key, **kwargs)

    def get_common_paths(self) -> dict:
        """Return the dictionary of common paths"""
        return self.name_factory.get_common_paths()

    def get_common_path(self, path_key: str, **kwargs: Any) -> str:
        """Resolve and return a common path using the kwargs as interopolants"""
        return self.name_factory.resolve_common_path(path_key, **kwargs)

    def get_files(self) -> dict[str, RailProjectFileTemplate]:
        """Return the dictionary of specific file templates"""
        if self._file_templates is not None:
            return self._file_templates
        if "all" in self.config.Files:
            self._file_templates = RailProjectFileFactory.get_file_templates()
        else:  # pragma: no cover
            self._file_templates = {
                key: RailProjectFileFactory.get_file_template(key)
                for key in self.config.Files
            }
        return self._file_templates

    def get_file(self, name: str, **kwargs: Any) -> str:
        """Resolve and return a file using the kwargs as interpolants"""
        files = self.get_files()
        try:
            file_template = files[name]
        except KeyError as missing_key:
            raise KeyError(
                f"file '{name}' not found in {list(files.keys())}"
            ) from missing_key

        path = self.name_factory.resolve_path(
            file_template.config.to_dict(), "path_template", **kwargs
        )
        return path

    def get_flavors(self) -> dict[str, RailFlavor]:
        """Return the dictionary of analysis flavor variants"""
        if self._flavors is not None:
            return self._flavors

        baseline = self.config.Baseline
        self._flavors = dict(baseline=RailFlavor(name="baseline", **baseline))

        flavor_list = self.config.Flavors

        for flavor_item in flavor_list:
            flavor_dict = baseline.copy()
            flavor_dict.update(**flavor_item["Flavor"])
            flavor_name = flavor_dict["name"]
            self._flavors[flavor_name] = RailFlavor(**flavor_dict)

        return self._flavors

    def get_flavor(self, name: str) -> RailFlavor:
        """Resolve the configuration for a particular analysis flavor variant"""
        flavors = self.get_flavors()
        try:
            return flavors[name]
        except KeyError as missing_key:
            raise KeyError(
                f"flavor '{name}' not found in {list(flavors.keys())}"
            ) from missing_key

    def get_file_for_flavor(self, flavor: str, label: str, **kwargs: Any) -> str:
        """Resolve the file associated to a particular flavor and label

        E.g., flavor=baseline and label=train would give the baseline training file
        """
        flavor_dict = self.get_flavor(flavor)
        try:
            file_alias = flavor_dict.config.file_aliases[label]
        except KeyError as msg:
            raise KeyError(f"Label '{label}' not found in flavor '{flavor}'") from msg
        return self.get_file(file_alias, flavor=flavor, label=label, **kwargs)

    def get_file_metadata_for_flavor(
        self, flavor: str, label: str
    ) -> RailProjectFileTemplate:
        """Resolve the metadata associated to a particular flavor and label

        E.g., flavor=baseline and label=train would give the baseline training metadata
        """
        flavor_dict = self.get_flavor(flavor)
        try:
            file_alias = flavor_dict["file_aliases"][label]
        except KeyError as msg:
            raise KeyError(f"Label '{label}' not found in flavor '{flavor}'") from msg
        return self.get_files()[file_alias]

    def get_selections(self) -> dict[str, RailSelection]:
        """Get the dictionary describing all the selections"""
        if self._selections is not None:
            return self._selections
        sel_names = (
            RailSelectionFactory.get_selection_names()
            if "all" in self.config.Selections
            else self.config.Selections
        )
        self._selections = {
            name_: RailSelectionFactory.get_selection(name_) for name_ in sel_names
        }
        return self._selections

    def get_selection(self, name: str) -> RailSelection:
        """Get a particular selection by name"""
        selections = self.get_selections()
        try:
            return selections[name]
        except KeyError as missing_key:
            raise KeyError(
                f"Selection '{name}' not found in {self}. "
                f"Known values are {list(selections.keys())}"
            ) from missing_key

    def get_subsamples(self) -> dict[str, RailSubsample]:
        """Get the dictionary describing all the subsamples"""
        if self._subsamples is not None:
            return self._subsamples
        sel_names = (
            RailSubsampleFactory.get_subsample_names()
            if "all" in self.config.Subsamples
            else self.config.Subsamples
        )
        self._subsamples = {
            name_: RailSubsampleFactory.get_subsample(name_) for name_ in sel_names
        }
        return self._subsamples

    def get_subsample(self, name: str) -> RailSubsample:
        """Get a particular subsample by name"""
        subsamples = self.get_subsamples()
        try:
            return subsamples[name]
        except KeyError as missing_key:
            raise KeyError(
                f"Subsample '{name}' not found in {self}. "
                f"Known values are {list(subsamples.keys())}"
            ) from missing_key

    def get_algorithms(self, algorithm_type: str) -> dict[str, dict[str, str]]:
        """Get all the algorithms of a particular type"""
        sub_algo_dict = self._algorithms.get(algorithm_type, {})
        if sub_algo_dict:
            return sub_algo_dict
        algo_names = self.config[algorithm_type]
        # trim off the trailing 's'
        all_algos = RailAlgorithmFactory.get_algorithms(algorithm_type[0:-1])
        use_algos = (
            list(all_algos.values())
            if "all" in algo_names
            else [all_algos[key] for key in algo_names]
        )
        for algo_ in use_algos:
            algo_.fill_dict(sub_algo_dict)
        self._algorithms[algorithm_type] = sub_algo_dict
        return sub_algo_dict

    def get_algorithm(self, algorithm_type: str, algo_name: str) -> dict[str, str]:
        """Get an algorithm of a particular type with a specific name"""
        algo_dict = self.get_algorithms(algorithm_type)
        try:
            return algo_dict[algo_name]
        except KeyError as missing_key:
            raise KeyError(
                f"Algorithm '{algo_name}' of type '{algorithm_type}' not found in {self}. "
                f"Known values are {list(algo_dict.keys())}"
            ) from missing_key

    def get_error_models(self) -> dict:
        """Get the dictionary describing all the photometric error model algorithms"""
        return self.get_algorithms("ErrorModels")

    def get_error_model(self, name: str) -> dict:
        """Get the information about a particular photometric error model algorithms"""
        return self.get_algorithm("ErrorModels", name)

    def get_pzalgorithms(self) -> dict:
        """Get the dictionary describing all the PZ estimation algorithms"""
        return self.get_algorithms("PZAlgorithms")

    def get_pzalgorithm(self, name: str) -> dict:
        """Get the information about a particular PZ estimation algorithm"""
        return self.get_algorithm("PZAlgorithms", name)

    def get_spec_selections(self) -> dict:
        """Get the dictionary describing all the spectroscopic selection algorithms"""
        return self.get_algorithms("SpecSelections")

    def get_spec_selection(self, name: str) -> dict:
        """Get the information about a particular spectroscopic selection algorithm"""
        return self.get_algorithm("SpecSelections", name)

    def get_classifiers(self) -> dict:
        """Get the dictionary describing all the tomographic bin classification"""
        return self.get_algorithms("Classifiers")

    def get_classifier(self, name: str) -> dict:
        """Get the information about a particular tomographic bin classification"""
        return self.get_algorithm("Classifiers", name)

    def get_summarizers(self) -> dict:
        """Get the dictionary describing all the NZ summarization algorithms"""
        return self.get_algorithms("Summarizers")

    def get_summarizer(self, name: str) -> dict:
        """Get the information about a particular NZ summarization algorithms"""
        return self.get_algorithm("Summarizers", name)

    def get_catalogs(self) -> dict:
        """Get the dictionary describing all the types of data catalogs"""
        if self._catalog_templates is not None:
            return self._catalog_templates

        self._catalog_templates = (
            RailCatalogFactory.get_catalog_templates()
            if "all" in self.config.Catalogs
            else {
                key: RailCatalogFactory.get_catalog_template(key)
                for key in self.config.Catalogs
            }
        )
        return self._catalog_templates

    def get_catalog_files(self, name: str, **kwargs: Any) -> list[str]:
        """Resolve the paths for a particular catalog file"""
        catalogs = self.get_catalogs()
        try:
            catalog = catalogs[name]
        except KeyError as missing_key:
            raise KeyError(
                f"catalogs '{name}' not found in {list(catalogs.keys())}"
            ) from missing_key
        interpolants = kwargs.copy()
        interpolants.update(**self.name_factory.get_resolved_common_paths())
        catalog_instance = catalog.make_catalog_instance("dummy", **interpolants)
        return catalog_instance.resolve(**self.config.IterationVars)

    def get_catalog(self, name: str, **kwargs: Any) -> str:
        """Resolve the path for a particular catalog file"""
        catalogs = self.get_catalogs()
        try:
            catalog = catalogs[name]
        except KeyError as missing_key:
            raise KeyError(
                f"catalogs '{name}' not found in {list(catalogs.keys())}"
            ) from missing_key

        try:
            path = self.name_factory.resolve_path(
                catalog.config.to_dict(), "path_template", **kwargs
            )
            return path
        except KeyError as missing_key:
            raise KeyError(f"path_template not found in {catalog}") from missing_key

    def get_pipelines(self) -> dict[str, RailPipelineTemplate]:
        """Get the dictionary describing all the types of ceci pipelines"""
        if self._pipeline_templates is not None:
            return self._pipeline_templates
        self._pipeline_templates = (
            RailPipelineFactory.get_pipeline_templates()
            if "all" in self.config.Pipelines
            else {
                key: RailPipelineFactory.get_pipeline_template(key)
                for key in self.config.Pipelines
            }
        )
        return self._pipeline_templates

    def get_pipeline(self, name: str) -> RailPipelineTemplate:
        """Get the information about a particular ceci pipeline"""
        pipelines = self.get_pipelines()
        try:
            return pipelines[name]
        except KeyError as missing_key:
            raise KeyError(
                f"pipeline '{name}' not found in {list(pipelines.keys())}"
            ) from missing_key

    def get_flavor_args(self, flavors: list[str]) -> list[str]:
        """Get the 'flavors' to iterate a particular command over

        Notes
        -----
        If the flavor 'all' is included in the list of flavors, this
        will replace the list with all the flavors defined in this project
        """
        flavor_dict = self.get_flavors()
        if "all" in flavors:
            return list(flavor_dict.keys())
        return flavors

    def get_selection_args(self, selections: list[str]) -> list[str]:
        """Get the 'selections' to iterate a particular command over

        Notes
        -----
        If the selection 'all' is included in the list of selections, this
        will replace the list with all the selections defined in this project
        """
        selection_dict = self.get_selections()
        if "all" in selections:
            return list(selection_dict.keys())
        return selections

    def wrap_pz_model(self, path: str, outdir: str, **kwargs: Any) -> int:
        """Wrap a pz model file for use by Rubin DM software

        Parameters
        ----------
        path
            Path to the model file

        outdir
            Directory we are writing to

        Returns
        -------
        status: 0 for success, error_code otherwise
        """
        tokens = path.split("_")
        algo_name = tokens[-1].replace(".pkl", "")
        try:
            algo = self.get_algorithm("PZAlgorithms", algo_name)
        except KeyError:
            return 0
        selection = kwargs["selection"]
        flavor_name = kwargs["flavor"]
        flavor = self.get_flavor(flavor_name)
        module = algo["Module"]
        informer = algo["Inform"]
        catalog_tag = flavor.config.catalog_tag
        creation_class_name = f"{module}.{informer}"
        outpath = os.path.join(
            outdir, f"model_{self.name}_{algo_name}_{flavor}_{selection}.pickle"
        )
        Model.wrap(
            path,
            outpath,
            creation_class_name=creation_class_name,
            version=0,
            catalog_tag=catalog_tag,
            provenance=dict(
                project=self.name,
                flavor=flavor_name,
                selection=selection,
            ),
        )
        return 0



from __future__ import annotations

from typing import Any, TypeVar

T = TypeVar("T", bound="DynamicClass")


class DynamicClass:
    """Base class for classes that will create objects of sub-classes dynamically

    This implements:

    1. keeping track of all the loaded sub-classes,
    2. being able to load a new sub-class from the python module and class name
    3. being able to create an object of sub-class from a dict

    Subclasses that serve as parent classes to several addtional sub-classes should

    1. override the sub_classes to keep track of their sub-classes
    """

    sub_classes: dict[str, type[DynamicClass]] = {}

    def __init_subclass__(cls) -> None:
        cls.sub_classes[cls.__name__] = cls

    @classmethod
    def print_classes(cls) -> None:
        """Print the sub-classes of DynamicClass that have been loaded"""
        print(f"{cls.__name__}")
        for key, val in cls.sub_classes.items():
            print(f"  {key} {val}")

    @classmethod
    def get_sub_class(cls: type[T], key: str, class_name: str | None = None) -> type[T]:
        """Get a particular sub-class of DynamicClass by name

        Parameters
        ----------
        key: str
            Key for the subclass.

        class_name: str | None = None
           Full class name, so it can be loaded

        Returns
        -------
        type:
            Subclass in question
        """
        if key in cls.sub_classes:
            sub_class = cls.sub_classes[key]
            assert issubclass(sub_class, cls)
            return sub_class
        if class_name is None:
            raise KeyError(
                f"class {key} not found in {list(cls.sub_classes.keys())} and class_name is not provided."
            )
        return cls.load_sub_class(class_name)

    @classmethod
    def load_sub_class(cls: type[T], class_name: str) -> type[T]:
        """Import a particular sub-class of DynamicClass by name

        Parameters
        ----------
        class_name: str
            Full path and name of the subclass, e.g., rail.plotting.some_file.SomeClass

        Returns
        -------
        type:
            Subclass in question
        """
        tokens = class_name.split(".")
        module = ".".join(tokens[:-1])
        key = tokens[-1]
        __import__(module)
        sub_class = cls.get_sub_class(key)
        assert issubclass(sub_class, cls)
        return sub_class

    @classmethod
    def create_from_dict(
        cls: type[T],
        config_dict: dict[str, Any],
    ) -> T:
        """Create a DynamicClass object of type T

        Parameters
        ----------
        config_dict: dict[str, Any],
            Configuration parameters

        Returns
        -------
        T:
            Newly created object

        Notes
        -----
        config_dict must include 'class_name' which gives the path and name of the
        class, e.g., rail.plotters.some_file.SomeClass
        """
        copy_config = config_dict.copy()
        class_name = copy_config.pop("class_name")
        tokens = class_name.split(".")
        key = tokens[-1]
        sub_class = cls.get_sub_class(key, class_name)
        assert issubclass(sub_class, cls)
        return sub_class(**copy_config)



import os

import pytest

from rail.plotting.dataset_factory import RailDatasetFactory
from rail.plotting.dataset_holder import (
    RailDatasetHolder,
    RailDatasetListHolder,
    RailProjectHolder,
)


def test_load_yaml(setup_project_area: int) -> None:
    assert setup_project_area == 0

    # Load the testing yaml file
    RailDatasetFactory.clear()
    RailDatasetFactory.load_yaml("tests/ci_datasets.yaml")
    RailDatasetFactory.print_contents()
    RailDatasetHolder.print_classes()

    with pytest.raises(KeyError):
        RailDatasetHolder.get_sub_class("bad")

    # Make sure the names of the projects got loaded
    the_project_names = RailDatasetFactory.get_project_names()
    assert "ci_test" in the_project_names

    # Make sure the projects got loaded
    the_dict = RailDatasetFactory.get_projects()
    assert isinstance(the_dict["ci_test"], RailProjectHolder)

    a_project_holder = RailDatasetFactory.get_project("ci_test")
    assert isinstance(a_project_holder, RailProjectHolder)

    # Make sure the names of the datasets got loaded
    the_dataset_names = RailDatasetFactory.get_dataset_names()
    assert "blend_baseline_knn" in the_dataset_names

    # Make sure the datasets got loaded
    the_datasets = RailDatasetFactory.get_datasets()
    assert isinstance(the_datasets["blend_baseline_knn"], RailDatasetHolder)

    # get a specfic dataset
    the_dataset = RailDatasetFactory.get_dataset("blend_baseline_knn")
    assert isinstance(the_dataset, RailDatasetHolder)

    with pytest.raises(KeyError):
        RailDatasetFactory.get_dataset("bad")

    # get a specfic dataset dict
    the_dataset_list = RailDatasetFactory.get_dataset_list("baseline_test")
    assert isinstance(the_dataset_list, RailDatasetListHolder)

    with pytest.raises(KeyError):
        RailDatasetFactory.get_dataset_list("bad")

    # Make sure the names of the datasets lists got loaded
    the_dataset_list_names = RailDatasetFactory.get_dataset_list_names()
    assert "baseline_test" in the_dataset_list_names

    # Make sure the  datasets lists got loaded
    the_dataset_lists = RailDatasetFactory.get_dataset_lists()
    assert isinstance(the_dataset_lists["baseline_test"], RailDatasetListHolder)

    # Test the interactive stuff
    RailDatasetFactory.clear()

    RailDatasetFactory.add_project(a_project_holder)
    check_project_holder = RailDatasetFactory.get_project("ci_test")
    assert isinstance(check_project_holder, RailProjectHolder)

    RailDatasetFactory.add_dataset(the_dataset)
    RailDatasetFactory.add_dataset_list(
        RailDatasetListHolder(
            name="test_list",
            dataset_class=the_dataset.output_type.full_class_name(),
            datasets=[the_dataset.config.name],
        )
    )

    check_dataset = RailDatasetFactory.get_dataset("blend_baseline_knn")
    assert isinstance(check_dataset, type(the_dataset))

    check_list = RailDatasetFactory.get_dataset_list("test_list")
    assert the_dataset.config.name in check_list.config.datasets

    # check writing the yaml dict
    RailDatasetFactory.write_yaml("tests/temp.yaml")
    RailDatasetFactory.clear()
    RailDatasetFactory.load_yaml("tests/temp.yaml")
    os.unlink("tests/temp.yaml")

    check_dataset = RailDatasetFactory.get_dataset("blend_baseline_knn")
    assert isinstance(check_dataset, type(the_dataset))

    check_list = RailDatasetFactory.get_dataset_list("test_list")
    assert the_dataset.config.name in check_list.config.datasets



from __future__ import annotations
from typing import Any

import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.dataset as ds
from pyarrow.parquet import filters_to_expression


# Legal PyArrow comparison operators for filter expressions
LEGAL_OPERATIONS = ["==", "!=", "<", "<=", ">", ">=", "in", "not in"]


def check_list_as_tuple(the_list: list[Any]) -> bool:
    """
    Test if a list can be converted to a PyArrow filter tuple.

    A valid PyArrow filter tuple has the form: (column_name, operator, value)
    where column_name is a string, operator is one of the legal operations,
    and value is the comparison value.

    Parameters
    ----------
    the_list
        List to test for conversion to a filter tuple.

    Returns
    -------
        True if the list is a valid 3-element filter specification.

    Raises
    ------
    ValueError
        If the list has 3 elements but the operator (second element) is not
        a legal PyArrow operation.

    Examples
    --------
    >>> check_list_as_tuple(['age', '>', 25])
    True
    >>> check_list_as_tuple(['category', 'in', ['A', 'B']])
    True
    >>> check_list_as_tuple(['age', 'invalid_op', 25])
    ValueError: Expected three item list of the form ['key', 'operation', value]...
    >>> check_list_as_tuple(['age', '>'])
    False

    Notes
    -----
    This function returns False for lists that don't have exactly 3 elements
    or whose first element is not a string. It only raises ValueError when
    the structure is correct but the operator is invalid.
    """
    if len(the_list) != 3:
        return False
    if not isinstance(the_list[0], str):
        return False
    if the_list[1] not in LEGAL_OPERATIONS:
        msg = "Expected three item list of the form ['key', 'operation', value].  "
        msg += f"{the_list[1]} is not in {LEGAL_OPERATIONS}."
        raise ValueError(msg)
    return True


def parse_item(
    item: list[Any] | dict[str, Any], top: bool = True
) -> list[Any] | tuple[Any, ...]:
    """
    Parse a filter specification into PyArrow filter format.

    This function recursively processes nested filter specifications and
    converts them into the format expected by PyArrow's filters_to_expression
    function. It handles both AND logic (nested lists) and OR logic
    (via 'OR_GROUP' dictionaries).

    Parameters
    ----------
    item
        Filter specification to parse. Can be:

        - List with 3 elements: ['column', 'op', value] → filter tuple
        - Nested list: [[filter1], [filter2]] → OR groups
        - Dict with 'OR_GROUP' key: {'OR_GROUP': [...]} → OR logic
    top
        Whether this is a top-level call. Top-level 3-tuples are wrapped
        in a list for consistency with PyArrow's expected format.

    Returns
    -------
        Parsed filter specification:

        - Single filter: [('column', 'op', value)] (if top=True)
        - Single filter: ('column', 'op', value) (if top=False)
        - Nested filters: [[tuple, ...], [tuple, ...], ...] for OR groups
        - AND filters: [tuple, tuple, ...] for conditions in same group

    Raises
    ------
    ValueError
        If a 3-element list has an invalid operator.
    AssertionError
        If item is neither a list nor a dict, or if a dict doesn't contain
        an 'OR_GROUP' key.

    Examples
    --------
    Simple filter (top level):
    >>> parse_item(['age', '>', 25])
    [('age', '>', 25)]

    Simple filter (nested):
    >>> parse_item(['age', '>', 25], top=False)
    ('age', '>', 25)

    AND logic (multiple conditions):
    >>> parse_item([['age', '>', 25], ['category', '==', 'A']])
    [('age', '>', 25), ('category', '==', 'A')]

    OR logic using OR_GROUP:
    >>> parse_item({'OR_GROUP': [['age', '>', 50], ['category', '==', 'VIP']]})
    [('age', '>', 50), ('category', '==', 'VIP')]

    Nested OR groups:
    >>> parse_item([[['age', '>', 25]], [['category', '==', 'A']]])
    [[('age', '>', 25)], [('category', '==', 'A')]]

    Notes
    -----
    - AND logic: Multiple filters in the same list → all must match
    - OR logic: Multiple sublists or 'OR_GROUP' dict → any can match
    - The function is recursive and handles arbitrary nesting depth
    - Top-level single filters are wrapped in a list for PyArrow compatibility
    """
    if isinstance(item, list):
        # it is a list, test if is a 3-tuple
        if check_list_as_tuple(item):
            # it is a 3-tuple, convert it
            tt = tuple(item)
            if top:
                # at the top-level, we wrap it in a list
                return [tt]
            # return it as a 3-tuple
            return tt
        # recursively process items on list as a tuple
        return [parse_item(sub_item, False) for sub_item in item]
    if isinstance(item, dict):
        # it is a dict, make sure it just consists of 'OR_GROUP' item
        assert (
            "OR_GROUP" in item
        ), f"Dict must contain 'OR_GROUP' key, got {item.keys()}"
        sub_list = item["OR_GROUP"]
        return parse_item(sub_list, False)
    raise AssertionError(f"Item {item} is neither a list nor a dict")


def filter_dataset(
    dataset: ds.Dataset,
    filter_conditions: list[tuple[str, str, Any]] | list[list[tuple[str, str, Any]]],
    columns: list[str] | dict[str, str],
) -> _ProjectedDataset:
    """
    Filter a PyArrow dataset and select specific columns.

    This is a convenience wrapper around PyArrow's native filtering and
    column projection functionality.

    Parameters
    ----------
    dataset
        The PyArrow dataset to filter.
    filter_conditions
        Filter specification using PyArrow's standard filter format:

        - List of tuples: [(column, op, value), ...] for AND logic
        - List of lists: [[(col, op, val), ...], [(col, op, val), ...]]
          for OR logic between groups

        Supported operators: '==', '!=', '<', '<=', '>', '>=', 'in', 'not in'

        Examples:
            [('category', '==', 'A'), ('value', '>', 10)]  # AND
            [[('category', '==', 'A')], [('value', '>=', 50)]]  # OR
    columns
        Column specification for projection:

        - List of column names to retain
        - Dict mapping new names to original names for renaming

        Examples:
            ['id', 'name', 'value']
            {'user_id': 'id', 'user_name': 'name'}

    Returns
    -------
        Filtered dataset with column projection applied. The dataset remains
        lazy and will be materialized only when explicitly converted.

    Raises
    ------
    ValueError
        If filter conditions are invalid or columns don't exist in schema.

    Examples
    --------
    >>> import pyarrow as pa
    >>> import pyarrow.dataset as ds
    >>>
    >>> table = pa.table({
    ...     'id': [1, 2, 3, 4],
    ...     'category': ['A', 'B', 'A', 'C'],
    ...     'value': [10, 20, 30, 40]
    ... })
    >>> dataset = ds.dataset(table)

    Simple AND filter:
    >>> filtered = filter_dataset(
    ...     dataset,
    ...     [('category', '==', 'A'), ('value', '>', 15)],
    ...     ['id', 'value']
    ... )

    OR logic:
    >>> filtered = filter_dataset(
    ...     dataset,
    ...     [[('category', '==', 'A')], [('value', '>=', 40)]],
    ...     ['id', 'category', 'value']
    ... )

    Column renaming:
    >>> filtered = filter_dataset(
    ...     dataset,
    ...     [('category', '==', 'A')],
    ...     {'identifier': 'id', 'amount': 'value'}
    ... )

    Notes
    -----
    - Uses PyArrow's native `filters_to_expression` for filter construction
    - Column projection is applied during scanning, maintaining lazy evaluation
    - See PyArrow documentation for full filter syntax details
    """
    # Build filter expression using PyArrow's built-in function
    if filter_conditions:
        filter_expr = filters_to_expression(filter_conditions)
        filtered_dataset = dataset.filter(filter_expr)
    else:
        filtered_dataset = dataset

    # Validate columns exist in schema
    schema_names = set(filtered_dataset.schema.names)

    if isinstance(columns, dict):
        # Dict: {new_name: old_name}
        missing_cols = set(columns.values()) - schema_names
        if missing_cols:
            raise ValueError(f"Columns not found in dataset schema: {missing_cols}")
        column_spec: dict[str, str] | list[str] = columns
    else:
        # List: [col1, col2, ...]
        missing_cols = set(columns) - schema_names
        if missing_cols:
            raise ValueError(f"Columns not found in dataset schema: {missing_cols}")
        column_spec = columns

    # Return a wrapper that applies column projection during scanning
    return _ProjectedDataset(filtered_dataset, column_spec)


class _ProjectedDataset:
    """
    Wrapper for a dataset with column projection applied lazily.

    This class wraps a PyArrow dataset and applies column projection
    (selection and/or renaming) only when the dataset is scanned or
    materialized.
    """

    def __init__(
        self,
        dataset: ds.Dataset,
        columns: list[str] | dict[str, str],
    ):
        """
        Initialize the projected dataset.

        Parameters
        ----------
        dataset
            Source dataset to wrap.
        columns
            Column specification (list for selection, dict for renaming).
        """
        self._dataset = dataset
        self._columns = columns

    @property
    def schema(self) -> pa.Schema:
        """Get the schema of the underlying dataset."""
        return self._dataset.schema

    def __getattr__(self, name: str) -> Any:
        """Delegate attribute access to the underlying dataset."""
        return getattr(self._dataset, name)

    def scanner(self, **kwargs: Any) -> ds.Scanner:
        """Create a scanner with column projection applied."""
        if "columns" not in kwargs:
            if isinstance(self._columns, dict):
                # For dict, we need to use expressions
                kwargs["columns"] = {
                    new_name: pc.field(old_name)
                    for new_name, old_name in self._columns.items()
                }
            else:
                kwargs["columns"] = self._columns
        return self._dataset.scanner(**kwargs)

    def to_table(self, **kwargs: Any) -> pa.Table:
        """Materialize the dataset with column projection applied."""
        return self.scanner(**kwargs).to_table()


def inner_join_datasets(
    datasets: dict[str, ds.Dataset | _ProjectedDataset],
    join_key: str,
    use_threads: bool = True,
) -> ds.Dataset:
    """
    Perform an inner join on multiple datasets using a specified join key.

    This is a thin wrapper around PyArrow's native join functionality.
    Column name collisions are automatically handled by PyArrow using
    suffixes based on dataset names.

    Parameters
    ----------
    datasets
        Dictionary mapping dataset names to PyArrow datasets. Names are used
        as suffixes for conflicting column names. Must contain at least one
        dataset.
    join_key
        Name of the column to use as the join key. This column must exist in
        all datasets.
    use_threads
        Whether to use multiple threads for the join operation.

    Returns
    -------
        Materialized table containing the inner join result. The join key
        column appears only once. Conflicting column names are suffixed with
        dataset names (e.g., 'value_users', 'value_orders').

    Raises
    ------
    ValueError
        If no datasets are provided or if the join key is missing from any
        dataset.

    Examples
    --------
    >>> import pyarrow as pa
    >>> import pyarrow.dataset as ds
    >>>
    >>> users_ds = ds.dataset(pa.table({
    ...     'id': [1, 2, 3],
    ...     'name': ['Alice', 'Bob', 'Charlie'],
    ...     'age': [25, 30, 35]
    ... }))
    >>> orders_ds = ds.dataset(pa.table({
    ...     'id': [2, 3, 4],
    ...     'total': [100, 200, 150]
    ... }))
    >>> result = inner_join_datasets(
    ...     {'users': users_ds, 'orders': orders_ds},
    ...     'id'
    ... )
    >>> result.column_names
    ['id', 'name', 'age', 'total']

    With column conflicts:
    >>> users_ds = ds.dataset(pa.table({
    ...     'id': [1, 2, 3],
    ...     'value': [10, 20, 30]
    ... }))
    >>> orders_ds = ds.dataset(pa.table({
    ...     'id': [2, 3, 4],
    ...     'value': [100, 200, 150]
    ... }))
    >>> result = inner_join_datasets(
    ...     {'users': users_ds, 'orders': orders_ds},
    ...     'id'
    ... )
    >>> result.column_names
    ['id', 'value_users', 'value_orders']

    Notes
    -----
    - Uses PyArrow's native Table.join() method with automatic suffix handling
    - For multiple datasets, joins are performed sequentially
    - Column name conflicts are resolved using dataset names as suffixes
    """
    if not datasets:
        raise ValueError("At least one dataset must be provided")

    # Validate join key exists in all datasets
    for name, dataset in datasets.items():
        if join_key not in dataset.schema.names:
            raise ValueError(
                f"Join key '{join_key}' not found in dataset '{name}' schema"
            )

    dataset_items = list(datasets.items())

    if len(dataset_items) == 1:
        # Single dataset
        _, dataset = dataset_items[0]
        if isinstance(dataset, _ProjectedDataset):
            return ds.dataset(dataset.to_table())
        return dataset

    # Perform sequential joins using PyArrow's native join with suffixes
    first_name, first_dataset = dataset_items[0]
    if isinstance(first_dataset, _ProjectedDataset):
        result = ds.dataset(first_dataset.to_table())
    else:
        result = first_dataset

    for name, dataset in dataset_items[1:]:
        if isinstance(dataset, _ProjectedDataset):
            right_table = ds.dataset(dataset.to_table())
        else:
            right_table = dataset

        # Perform inner join with automatic suffix handling
        result = result.join(
            right_table,
            keys=join_key,
            join_type="inner",
            left_suffix=f"_{first_name}",
            right_suffix=f"_{name}",
            use_threads=use_threads,
        )

        # Update the "left" name for subsequent joins
        # (this accumulates previous dataset names)
        first_name = f"{first_name}+{name}"

    return result



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

from rail.core.factory_mixin import RailFactoryMixin

from rail.projects import RailProject

from .dataset_holder import RailDatasetHolder, RailDatasetListHolder, RailProjectHolder

if TYPE_CHECKING:
    from rail.core.configurable import Configurable

    C = TypeVar("C", bound="Configurable")


class RailDatasetFactory(RailFactoryMixin):
    """Factory class to make datasets

    Expected usage is that user will define a yaml file with the various
    datasets that they wish to use with the following example syntax:

    .. highlight:: yaml
    .. code-block:: yaml

      Data:
        - Project:
           name: some_project
           yaml_file: /path/to/rail_project_file
        - Dataset:
            name: gold_baseline_test
            class: rail.plotting.project_dataset_holder.RailProjectDatasetHolder
            extractor: rail.plotting.pz_data_extractor.PZPointEstimateDataExtractor
            project: some_project
            selection: gold
            flavor: baseline
            tag: test
            algos: ['all']
        - Dataset:
            name: blend_baseline_test
            class: rail.plotting.project_dataset_holder.RailProjectDatasetHolder
            exctractor: rail.plottings.pz_data_extractor.PZPointEstimateDataExtractor
            project: some_project
            selection: blend
            flavor: baseline
            tag: test
            algos: ['all']

    And group them into lists of dataset that can be run over particular types
    of data, using the following example syntax:

    .. highlight:: yaml
    .. code-block:: yaml

      Data:
        - DatasetList:
            name: baseline_test
            datasets:
              - gold_baseline_test
              - blend_baseline_test
    """

    yaml_tag: str = "Data"

    client_classes = [RailProjectHolder, RailDatasetHolder, RailDatasetListHolder]

    _instance: RailDatasetFactory | None = None

    def __init__(self) -> None:
        """C'tor, build an empty RailDatasetFactory"""
        RailFactoryMixin.__init__(self)
        self._projects = self.add_dict(RailProjectHolder)
        self._datasets = self.add_dict(RailDatasetHolder)
        self._dataset_lists = self.add_dict(RailDatasetListHolder)

    @classmethod
    def get_projects(cls) -> dict[str, RailProject]:
        """Return the dict of all the projects"""
        return cls.instance().projects

    @classmethod
    def get_project_names(cls) -> list[str]:
        """Return the names of all the projects"""
        return list(cls.instance().projects.keys())

    @classmethod
    def get_project(cls, key: str) -> RailProject:
        """Return a project by name"""
        return cls.instance().projects[key]

    @classmethod
    def get_datasets(cls) -> dict[str, RailDatasetHolder]:
        """Return the dict of all the datasets"""
        return cls.instance().datasets

    @classmethod
    def get_dataset_names(cls) -> list[str]:
        """Return the names of the datasets"""
        return list(cls.instance().datasets.keys())

    @classmethod
    def get_dataset_lists(cls) -> dict[str, RailDatasetListHolder]:
        """Return the dict of all the datasets"""
        return cls.instance().dataset_lists

    @classmethod
    def get_dataset_list_names(cls) -> list[str]:
        """Return the names of the dataset lists"""
        return list(cls.instance().dataset_lists.keys())

    @classmethod
    def get_dataset(cls, name: str) -> RailDatasetHolder:
        """Get dataset by it's assigned name

        Parameters
        ----------
        name: str
            Name of the dataset to return

        Returns
        -------
        dataset: dict
            Dataset in question
        """
        try:
            return cls.instance().datasets[name]
        except KeyError as msg:
            raise KeyError(
                f"Dataset named {name} not found in RailDatasetFactory "
                f"{list(cls.instance().datasets.keys())}"
            ) from msg

    @classmethod
    def get_dataset_list(cls, name: str) -> RailDatasetListHolder:
        """Get a list of datasets their assigned name

        Parameters
        ----------
        name: str
            Name of the dataset list to return

        Returns
        -------
        datasets: list[dict]
            Datasets in question
        """
        try:
            return cls.instance().dataset_lists[name]
        except KeyError as msg:
            raise KeyError(
                f"DatasetList named {name} not found in RailDatasetFactory "
                f"{list(cls.instance().dataset_lists.keys())}"
            ) from msg

    @classmethod
    def add_project(cls, project_holder: RailProjectHolder) -> None:
        """Add a particular RailProjectHolder to the factory"""
        cls.instance().add_to_dict(project_holder)

    @classmethod
    def add_dataset(cls, dataset_holder: RailDatasetHolder) -> None:
        """Add a particular RailDatasetHolder to the factory"""
        cls.instance().add_to_dict(dataset_holder)

    @classmethod
    def add_dataset_list(cls, dataset_list: RailDatasetListHolder) -> None:
        """Add a particular RailDatasetListHolder to the factory"""
        cls.instance().add_to_dict(dataset_list)

    @property
    def projects(self) -> dict[str, RailProjectHolder]:
        """Return the dictionary of RailProjects"""
        return self._projects

    @property
    def datasets(self) -> dict[str, RailDatasetHolder]:
        """Return the dictionary of individual datasets"""
        return self._datasets

    @property
    def dataset_lists(self) -> dict[str, RailDatasetListHolder]:
        """Return the dictionary of lists of datasets"""
        return self._dataset_lists

    def load_object_from_yaml_tag(
        self, configurable_class: type[C], yaml_tag: dict[str, Any]
    ) -> None:
        if configurable_class == RailDatasetHolder:
            the_object = RailDatasetHolder.create_from_dict(yaml_tag)
            self.add_to_dict(the_object)
            the_object.resolve()
            return
        RailFactoryMixin.load_object_from_yaml_tag(self, configurable_class, yaml_tag)



import pytest

from rail.projects import library


@pytest.fixture(name="setup_project_area", scope="package")
def setup_project_area(request: pytest.FixtureRequest) -> int:
    ret_val = library.setup_project_area()

    request.addfinalizer(library.teardown_project_area)

    return ret_val



from __future__ import annotations

import itertools
import os
from typing import Any

from ceci.config import StageParameter
from rail.core.configurable import Configurable


class RailProjectCatalogInstance(Configurable):
    """Simple class for holding information need to make a coherent catalog
    of files using a templated file name and iteration_vars to fill in
    the interpolation in the file name.

    For example the path_template might be 'a_file/{healpix}/data.parqut'
    and the interation_vars would be ['healpix'].

    When called with a dict such as healpix : [3433, 3344] it would the
    path_template would get expanded out to two files:

    a_file/3433/data.parqut
    a_file/3344/data.parqut

    """

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Dataset name"),
        path_template=StageParameter(
            str, None, fmt="%s", required=True, msg="Template for path to catalog files"
        ),
        iteration_vars=StageParameter(
            list,
            [],
            fmt="%s",
            msg="Variables to iterate over to construct catalog",
        ),
    )
    yaml_tag = "CatalogInstance"

    def __init__(self, **kwargs: Any):
        """C'tor

        Parameters
        ----------
        **kwargs: Any
            Configuration parameters for this RailProjectCatalogInstance, must match
            class.config_options data members
        """
        Configurable.__init__(self, **kwargs)
        self._file_list: list[str] | None = None
        self._file_exists: list[bool] | None = None

    def __repr__(self) -> str:
        return f"{self.config.path_template}"

    def resolve(self, **kwargs: dict[str, Any]) -> list[str]:
        """Resolve the list of files in this catalog

        :meta public:

        Parameters
        ----------
        **kwargs:
            Set of interpolants and iteration_vars needed to resolve the catalog

        Returns
        -------
        list[str]:
            List of resolved catalog files

        Notes
        -----
        By default this will used cached values, to override this and force rechecking
        use update=True keyword argument
        """
        update = kwargs.pop("update", False)
        if self._file_list is not None:
            if not update:
                return self._file_list
        iterations = itertools.product(*[kwargs.get(key, []) for key in kwargs])
        self._file_list = []
        for iteration_args in iterations:
            zipped_tuples = zip(self.config.iteration_vars, iteration_args)
            iteration_kwargs = {val_[0]: val_[1] for val_ in zipped_tuples}
            self._file_list.append(self.config.path_template.format(**iteration_kwargs))
        return self._file_list

    def check_files(self, **kwargs: dict[str, Any]) -> list[bool]:
        """Check if the files in the catalog exist

        Parameters
        ----------
        **kwargs:
            Set of interpolants and iteration_vars needed to resolve the catalog

        Returns
        -------
        list[bool]:
            List of True/False values for existance of each file in catalog

        Notes
        -----
        By default this will used cached values, to override this and force rechecking
        use update=True keyword argument
        """
        update = kwargs.get("update", False)
        if self._file_exists is not None:
            if not update:
                return self._file_exists
        self._file_exists = []
        the_files = self.resolve(**kwargs)
        for file_ in the_files:
            self._file_exists.append(os.path.exists(os.path.expandvars(file_)))
        return self._file_exists


class RailProjectCatalogTemplate(Configurable):
    """Simple class for holding a template for a catalog associated with a project

    The makes a coherent catalog of files using a templated file name, interpolants,
    and iteration_vars to fill in the interpolation in the file name.

    For example the path_template might be 'a_file/{healpix}/{flavor}_data.hdf5'
    and the interpolants would be ['flavor'] and interation_vars would be ['healpix'].

    When called with a dict such as flavor: 'baseline, healpix : [3433, 3344] it would the
    path_template would get expanded out to two files:

    a_file/3433/baseline_data.hdf5
    a_file/3344/baseline_data.hdf5
    """

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="Dataset name"),
        path_template=StageParameter(
            str, None, fmt="%s", required=True, msg="Template for path to catalog files"
        ),
        iteration_vars=StageParameter(
            list,
            [],
            fmt="%s",
            msg="Variables to iterate over to construct catalog",
        ),
    )
    yaml_tag = "CatalogTemplate"

    def __init__(self, **kwargs: Any) -> None:
        """C'tor

        Parameters
        ----------
        **kwargs: Any
            Configuration parameters for this RailProjectCatalogTemplate, must match
            class.config_options data members
        """
        Configurable.__init__(self, **kwargs)

    def __repr__(self) -> str:
        return f"{self.config.path_template}"

    def make_catalog_instance(
        self, name: str, **kwargs: dict[str, Any]
    ) -> RailProjectCatalogInstance:
        """Make and return a specific instance of this CatalogTemplate
        by resolving interpolants and iterating over the iteration_vars.

        Parameters
        ----------
        name:
            Name for the CatalogInstance object

        **kwargs:
            Interpolants needed to resolve the path template

        Returns
        -------
        RailProjectCatalogInstance:
            Newly created object
        """
        iteration_var_dict = {
            key: "{" + key + "}" for key in self.config.iteration_vars
        }
        formatted_path = self.config.path_template.format(
            **kwargs, **iteration_var_dict
        )
        return RailProjectCatalogInstance(
            name=name,
            path_template=formatted_path,
            iteration_vars=self.config.iteration_vars.copy(),
        )



from __future__ import annotations

import os
from typing import Any

import numpy as np
from astropy.stats import biweight_location, biweight_scale
from ceci.config import StageParameter
from matplotlib import colors
from matplotlib import pyplot as plt
from scipy.stats import sigmaclip

from .dataset import RailDataset
from .dataset_holder import RailDatasetHolder
from .plot_holder import RailPlotHolder
from .plotter import RailPlotter


class RailPZPointEstimateDataset(RailDataset):
    """Dataet to hold a vector p(z) point estimates and corresponding
    true redshifts
    """

    data_types = dict(truth=np.ndarray, pointEstimate=np.ndarray, magnitude=np.ndarray)


class RailPZMultiPointEstimateDataset(RailDataset):
    """Dataet to hold a set of vectors of p(z) point estimates and corresponding
    true redshifts
    """

    data_types = dict(
        truth=np.ndarray,
        pointEstimates=dict[str, np.ndarray],
    )


class PZPlotterPointEstimateVsTrueHist2D(RailPlotter):
    """Class to make a 2D histogram of p(z) point estimates
    versus true redshift
    """

    config_options: dict[str, StageParameter] = RailPlotter.config_options.copy()
    config_options.update(
        z_min=StageParameter(float, 0.0, fmt="%0.2f", msg="Minimum Redshift"),
        z_max=StageParameter(float, 3.0, fmt="%0.2f", msg="Maximum Redshift"),
        n_zbins=StageParameter(int, 150, fmt="%i", msg="Number of z bins"),
        n_clip=StageParameter(
            int, 3, fmt="%i", msg="Number of sigma cliping for outliers"
        ),
        abs_out_thresh=StageParameter(
            float, 0.2, fmt="%0.2f", msg="Threshold for the absolute outlier rate"
        ),
    )

    input_type = RailPZPointEstimateDataset

    def _make_2d_hist_plot(
        self,
        prefix: str,
        truth: np.ndarray,
        pointEstimate: np.ndarray,
        dataset_holder: RailDatasetHolder | None = None,
    ) -> RailPlotHolder:
        figure, axes = plt.subplots(figsize=(7, 6))
        bin_edges = np.linspace(
            self.config.z_min, self.config.z_max, self.config.n_zbins + 1
        )
        dz = (pointEstimate - truth) / (1 + truth)
        mean, _mean_err, std, outlier_rate, abs_outlier_rate = (
            self.get_biweight_mean_sigma_outlier(dz, nclip=self.config.n_clip)
        )
        mean, std, outlier_rate, abs_outlier_rate = (
            round(mean, 4),
            round(std, 4),
            round(outlier_rate, 4),
            round(abs_outlier_rate, 4),
        )
        h = axes.hist2d(
            truth,
            pointEstimate,
            bins=(bin_edges, bin_edges),
            norm=colors.LogNorm(),
            cmap="gray",
        )
        axes.plot(
            [self.config.z_min - 10, self.config.z_max + 10],
            [self.config.z_min - 10, self.config.z_max + 10],
            "--",
            color="red",
        )
        axes.plot(
            [self.config.z_min - 10, self.config.z_max + 10],
            [self.config.z_min - 10 - 3 * std, self.config.z_max + 10 - 3 * std],
            "--",
            color="red",
        )
        axes.plot(
            [self.config.z_min - 10, self.config.z_max + 10],
            [self.config.z_min - 10 + 3 * std, self.config.z_max + 10 + 3 * std],
            "--",
            color="red",
        )
        axes.plot(
            [],
            [],
            ".",
            alpha=0.0,
            label=rf"$\Delta z = {mean} $"
            + "\n"
            + rf"$\sigma z = {std} $"
            + "\n"
            + rf"outlier rate (>3$\sigma$) = {outlier_rate}"
            + "\n"
            + f"outlier rate (>{self.config.abs_out_thresh}) = {abs_outlier_rate}",
        )

        plt.xlabel("True Redshift")
        plt.ylabel("Estimated Redshift")
        cb = figure.colorbar(h[3], ax=axes)
        cb.set_label("Density")

        plt.legend()

        plot_name = self._make_full_plot_name(prefix, "")

        return RailPlotHolder(
            name=plot_name, figure=figure, plotter=self, dataset_holder=dataset_holder
        )

    def _make_plots(self, prefix: str, **kwargs: Any) -> dict[str, RailPlotHolder]:
        find_only = kwargs.get("find_only", False)
        figtype = kwargs.get("figtype", "png")
        dataset_holder = kwargs.get("dataset_holder")
        out_dict: dict[str, RailPlotHolder] = {}
        truth: np.ndarray = kwargs["truth"]
        pointEstimate: np.ndarray = kwargs["pointEstimate"]
        if find_only:
            plot_name = self._make_full_plot_name(prefix, "")
            assert dataset_holder
            plot = RailPlotHolder(
                name=plot_name,
                path=os.path.join(dataset_holder.config.name, f"{plot_name}.{figtype}"),
                plotter=self,
                dataset_holder=dataset_holder,
            )
        else:
            plot = self._make_2d_hist_plot(
                prefix=prefix,
                truth=truth,
                pointEstimate=pointEstimate,
                dataset_holder=dataset_holder,
            )
        out_dict[plot.name] = plot
        return out_dict

    def get_biweight_mean_sigma_outlier(
        self, subset: np.ndarray, nclip: int = 3
    ) -> tuple[float, float, float, float, float]:
        subset_clip, _, _ = sigmaclip(subset, low=3, high=3)
        for _j in range(nclip):
            subset_clip, _, _ = sigmaclip(subset_clip, low=3, high=3)

        mean = biweight_location(subset_clip)
        std = biweight_scale(subset_clip)
        outlier_rate = np.sum(np.abs(subset) > 3 * biweight_scale(subset_clip)) / len(
            subset
        )
        abs_outlier_rate = np.sum(np.abs(subset) > self.config.abs_out_thresh) / len(
            subset
        )

        return (
            mean,
            std / np.sqrt(len(subset_clip)),
            std,
            outlier_rate,
            abs_outlier_rate,
        )


class PZPlotterPointEstimateVsTrueProfile(RailPlotter):
    """Class to make a profile plot of p(z) point estimates
    versus true redshift
    """

    config_options: dict[str, StageParameter] = RailPlotter.config_options.copy()
    config_options.update(
        z_min=StageParameter(float, 0.0, fmt="%0.2f", msg="Minimum Redshift"),
        z_max=StageParameter(float, 3.0, fmt="%0.2f", msg="Maximum Redshift"),
        n_zbins=StageParameter(int, 150, fmt="%i", msg="Number of z bins"),
    )

    input_type = RailPZPointEstimateDataset

    def _make_2d_profile_plot(
        self,
        prefix: str,
        truth: np.ndarray,
        pointEstimate: np.ndarray,
        dataset_holder: RailDatasetHolder | None = None,
    ) -> RailPlotHolder:
        figure, axes = plt.subplots()
        bin_edges = np.linspace(
            self.config.z_min, self.config.z_max, self.config.n_zbins + 1
        )
        bin_centers = 0.5 * (bin_edges[0:-1] + bin_edges[1:])
        z_true_bin = np.searchsorted(bin_edges, truth)
        means = np.zeros((self.config.n_zbins))
        stds = np.zeros((self.config.n_zbins))
        for i in range(self.config.n_zbins):
            mask = z_true_bin == i
            data = pointEstimate[mask]
            if len(data) == 0:
                continue
            means[i] = np.mean(data) - bin_centers[i]
            stds[i] = np.std(data)

        axes.errorbar(
            bin_centers,
            means,
            stds,
        )
        plt.xlabel("True Redshift")
        plt.ylabel("Estimated Redshift")
        plot_name = self._make_full_plot_name(prefix, "")
        return RailPlotHolder(
            name=plot_name, figure=figure, plotter=self, dataset_holder=dataset_holder
        )

    def _make_plots(self, prefix: str, **kwargs: Any) -> dict[str, RailPlotHolder]:
        find_only = kwargs.get("find_only", False)
        figtype = kwargs.get("figtype", "png")
        dataset_holder = kwargs.get("dataset_holder")
        out_dict: dict[str, RailPlotHolder] = {}
        truth: np.ndarray = kwargs["truth"]
        pointEstimate: np.ndarray = kwargs["pointEstimate"]
        if find_only:
            assert dataset_holder
            plot_name = self._make_full_plot_name(prefix, "")
            plot = RailPlotHolder(
                name=plot_name,
                path=os.path.join(dataset_holder.config.name, f"{plot_name}.{figtype}"),
                plotter=self,
                dataset_holder=dataset_holder,
            )
        else:
            plot = self._make_2d_profile_plot(
                prefix=prefix,
                truth=truth,
                pointEstimate=pointEstimate,
                dataset_holder=dataset_holder,
            )
        out_dict[plot.name] = plot
        return out_dict


class PZPlotterAccuraciesVsTrue(RailPlotter):
    """Class to make a plot of the accuracy of several algorithms
    versus true redshift
    """

    config_options: dict[str, StageParameter] = RailPlotter.config_options.copy()
    config_options.update(
        z_min=StageParameter(float, 0.0, fmt="%0.2f", msg="Minimum Redshift"),
        z_max=StageParameter(float, 3.0, fmt="%0.2f", msg="Maximum Redshift"),
        n_zbins=StageParameter(int, 150, fmt="%i", msg="Number of z bins"),
        delta_cutoff=StageParameter(
            float, 0.1, fmt="%0.2f", msg="Delta-Z Cutoff for accurary"
        ),
    )

    input_type = RailPZMultiPointEstimateDataset

    def _make_accuracy_plot(
        self,
        prefix: str,
        truth: np.ndarray,
        pointEstimates: dict[str, np.ndarray],
        dataset_holder: RailDatasetHolder | None = None,
    ) -> RailPlotHolder:
        figure, axes = plt.subplots()
        bin_edges = np.linspace(
            self.config.z_min, self.config.z_max, self.config.n_zbins + 1
        )
        bin_centers = 0.5 * (bin_edges[0:-1] + bin_edges[1:])
        z_true_bin = np.searchsorted(bin_edges, truth)
        for key, val in pointEstimates.items():
            deltas = val - truth
            accuracy = np.ones((self.config.n_zbins)) * np.nan
            for i in range(self.config.n_zbins):
                mask = z_true_bin == i
                data = deltas[mask]
                if len(data) == 0:
                    continue
                accuracy[i] = (np.abs(data) <= self.config.delta_cutoff).sum() / float(
                    len(data)
                )
            axes.plot(
                bin_centers,
                accuracy,
                label=key,
            )
        plt.xlabel("True Redshift")
        plt.ylabel("Estimated Redshift")
        plot_name = self._make_full_plot_name(prefix, "")
        return RailPlotHolder(
            name=plot_name, figure=figure, plotter=self, dataset_holder=dataset_holder
        )

    def _make_plots(self, prefix: str, **kwargs: Any) -> dict[str, RailPlotHolder]:
        find_only = kwargs.get("find_only", False)
        figtype = kwargs.get("figtype", "png")
        dataset_holder = kwargs.get("dataset_holder")
        out_dict: dict[str, RailPlotHolder] = {}
        if find_only:
            plot_name = self._make_full_plot_name(prefix, "")
            assert dataset_holder
            plot = RailPlotHolder(
                name=plot_name,
                path=os.path.join(dataset_holder.config.name, f"{plot_name}.{figtype}"),
                plotter=self,
                dataset_holder=dataset_holder,
            )
        else:
            plot = self._make_accuracy_plot(prefix=prefix, **kwargs)
        out_dict[plot.name] = plot
        return out_dict


class PZPlotterBiweightStatsVsRedshift(RailPlotter):
    """Class to make a 2D histogram of p(z) point estimates
    versus true redshift
    """

    config_options: dict[str, StageParameter] = RailPlotter.config_options.copy()
    config_options.update(
        z_min=StageParameter(float, 0.0, fmt="%0.2f", msg="Minimum Redshift"),
        z_max=StageParameter(float, 3.0, fmt="%0.2f", msg="Maximum Redshift"),
        n_zbins=StageParameter(int, 15, fmt="%i", msg="Number of z bins"),
        n_clip=StageParameter(
            int, 3, fmt="%i", msg="Number of sigma cliping for outliers"
        ),
        zbin_type=StageParameter(
            str, "spec", fmt="%s", msg="Type of redshift binned by, 'spec' or 'phot'. "
        ),
    )

    input_type = RailPZPointEstimateDataset

    def _make_biweight_stats_plot(
        self,
        prefix: str,
        truth: np.ndarray,
        pointEstimate: np.ndarray,
        dataset_holder: RailDatasetHolder | None = None,
    ) -> RailPlotHolder:
        dz = (pointEstimate - truth) / (1 + truth)

        if self.config.zbin_type == "spec":
            x_label = r"$z_{spec}$"
            z_x = truth
        elif self.config.zbin_type == "phot":  # pragma: no cover
            x_label = r"$z_{phot}$"
            z_x = pointEstimate
        else:  # pragma: no cover
            raise ValueError("`zbin_type` must be either 'spec' or 'phot'")

        results = self.process_data(
            pointEstimate,
            truth,
            nbin=self.config.n_zbins,
            low=self.config.z_min,
            high=self.config.z_max,
            nclip=self.config.n_clip,
        )
        figure, axes = plt.subplots(2, 1, figsize=(8, 6))

        plt.subplots_adjust(wspace=0.1, hspace=0.0)

        axes[0].errorbar(
            results["z_mean"],
            results["biweight_mean"],
            results["biweight_std"],
            label="Bias",
        )

        axes[0].plot(results["z_mean"], results["biweight_sigma"], label=r"$\sigma_z$")

        axes[0].plot(
            results["z_mean"], results["biweight_outlier"], label=r"Outlier rate"
        )
        axes[0].set_title(
            f"Bias, Sigma, and Outlier rates w/ {self.config.n_clip} sigma clipping"
        )
        axes[0].set_ylabel("Statistics")
        axes[0].legend()
        axes[0].tick_params(
            axis="x", which="both", bottom=False, top=False, labelbottom=False
        )
        axes[0].set_xlim(self.config.z_min, self.config.z_max)

        bin_edges_z = np.linspace(self.config.z_min, self.config.z_max, 100 + 1)
        bin_edges_dz = np.linspace(np.min(dz), np.max(dz), 100 + 1)

        axes[1].hist2d(
            z_x,
            dz,
            bins=(bin_edges_z, bin_edges_dz),
            norm=colors.LogNorm(),
            cmap="gray",
        )

        axes[1].set_xlim(self.config.z_min, self.config.z_max)
        for qt in ["qt_95_low", "qt_68_low", "median", "qt_68_high", "qt_95_high"]:
            axes[1].plot(
                results["z_mean"], results[qt], "--", color="blue", linewidth=2.0
            )

        axes[1].set_xlabel(x_label)
        axes[1].set_ylabel(r"$(z_{phot} - z_{spec})/(1+z_{spec})$")
        plot_name = self._make_full_plot_name(prefix, "")
        return RailPlotHolder(
            name=plot_name, figure=figure, plotter=self, dataset_holder=dataset_holder
        )

    def _make_plots(self, prefix: str, **kwargs: Any) -> dict[str, RailPlotHolder]:
        find_only = kwargs.get("find_only", False)
        figtype = kwargs.get("figtype", "png")
        dataset_holder = kwargs.get("dataset_holder")
        out_dict: dict[str, RailPlotHolder] = {}
        truth: np.ndarray = kwargs["truth"]
        pointEstimate: np.ndarray = kwargs["pointEstimate"]
        if find_only:
            plot_name = self._make_full_plot_name(prefix, "")
            assert dataset_holder
            plot = RailPlotHolder(
                name=plot_name,
                path=os.path.join(dataset_holder.config.name, f"{plot_name}.{figtype}"),
                plotter=self,
                dataset_holder=dataset_holder,
            )
        else:
            plot = self._make_biweight_stats_plot(
                prefix=prefix,
                truth=truth,
                pointEstimate=pointEstimate,
                dataset_holder=dataset_holder,
            )
        out_dict[plot.name] = plot
        return out_dict

    def process_data(
        self,
        zphot: np.ndarray,
        specz: np.ndarray,
        low: float = 0.01,
        high: float = 2.0,
        nclip: int = 3,
        nbin: int = 101,
    ) -> dict[str, list[float]]:
        dz = (zphot - specz) / (1 + specz)

        z_bins = np.linspace(low, high, nbin)
        # Bin the data
        if self.config.zbin_type == "spec":
            zx = specz
        else:  # pragma: no cover
            zx = zphot

        bin_indices = np.digitize(zx, bins=z_bins) - 1  # Assign each point to a bin
        biweight_mean: list[float] = []
        biweight_std: list[float] = []
        biweight_sigma: list[float] = []
        biweight_outlier: list[float] = []
        z_mean: list[float] = []
        qt_95_low: list[float] = []
        qt_68_low: list[float] = []
        median: list[float] = []
        qt_68_high: list[float] = []
        qt_95_high: list[float] = []
        for i in range(len(z_bins) - 1):
            subset = dz[bin_indices == i]
            if len(subset) < 1:  # pragma: no cover
                continue
            subset_clip, _, _ = sigmaclip(subset, low=3, high=3)
            for _j in range(nclip):
                subset_clip, _, _ = sigmaclip(subset_clip, low=3, high=3)

            biweight_mean.append(biweight_location(subset_clip))
            biweight_std.append(biweight_scale(subset_clip) / np.sqrt(len(subset_clip)))
            biweight_sigma.append(biweight_scale(subset_clip))

            outlier_rate = np.sum(
                np.abs(subset) > 3 * biweight_scale(subset_clip)
            ) / len(subset)
            biweight_outlier.append(outlier_rate)

            qt_95_low.append(np.percentile(subset, 2.5))
            qt_68_low.append(np.percentile(subset, 16))
            median.append(np.percentile(subset, 50))
            qt_68_high.append(np.percentile(subset, 84))
            qt_95_high.append(np.percentile(subset, 97.5))

            z_mean.append(np.mean(zx[bin_indices == i]))

        return {
            "z_mean": z_mean,
            "biweight_mean": biweight_mean,
            "biweight_std": biweight_std,
            "biweight_sigma": biweight_sigma,
            "biweight_outlier": biweight_outlier,
            "qt_95_low": qt_95_low,
            "qt_68_low": qt_68_low,
            "median": median,
            "qt_68_high": qt_68_high,
            "qt_95_high": qt_95_high,
        }


class PZPlotterBiweightStatsVsMag(RailPlotter):
    """Class to make a 2D histogram of p(z) point estimates
    versus true redshift
    """

    config_options: dict[str, StageParameter] = RailPlotter.config_options.copy()
    config_options.update(
        mag_min=StageParameter(float, 18, fmt="%0.2f", msg="Minimum Magnitude"),
        mag_max=StageParameter(float, 25, fmt="%0.2f", msg="Maximum Magnitude"),
        n_magbins=StageParameter(int, 10, fmt="%i", msg="Number of magnitude bins"),
        n_clip=StageParameter(
            int, 3, fmt="%i", msg="Number of sigma cliping for outliers"
        ),
    )

    input_type = RailPZPointEstimateDataset

    def _make_biweight_stats_plot(
        self,
        prefix: str,
        truth: np.ndarray,
        pointEstimate: np.ndarray,
        magnitude: np.ndarray,
        dataset_holder: RailDatasetHolder | None = None,
    ) -> RailPlotHolder:
        dz = (pointEstimate - truth) / (1 + truth)

        results = self.process_data(
            pointEstimate,
            truth,
            magnitude,
            nbin=self.config.n_magbins,
            low=self.config.mag_min,
            high=self.config.mag_max,
            nclip=self.config.n_clip,
        )
        figure, axes = plt.subplots(2, 1, figsize=(8, 6))

        plt.subplots_adjust(wspace=0.1, hspace=0.0)

        axes[0].errorbar(
            results["mag_mean"],
            results["biweight_mean"],
            results["biweight_std"],
            label="Bias",
        )

        axes[0].plot(
            results["mag_mean"], results["biweight_sigma"], label=r"$\sigma_z$"
        )

        axes[0].plot(
            results["mag_mean"], results["biweight_outlier"], label=r"Outlier rate"
        )
        axes[0].set_title(
            f"Bias, Sigma, and Outlier rates w/ {self.config.n_clip} sigma clipping"
        )
        axes[0].set_ylabel("Statistics")
        axes[0].legend()
        axes[0].tick_params(
            axis="x", which="both", bottom=False, top=False, labelbottom=False
        )
        axes[0].set_xlim(self.config.mag_min, self.config.mag_max)

        bin_edges_mag = np.linspace(self.config.mag_min, self.config.mag_max, 100 + 1)
        bin_edges_dz = np.linspace(np.min(dz), np.max(dz), 100 + 1)
        axes[1].hist2d(
            magnitude,
            dz,
            bins=(bin_edges_mag, bin_edges_dz),
            norm=colors.LogNorm(),
            cmap="gray",
        )

        axes[1].set_xlim(self.config.mag_min, self.config.mag_max)
        for qt in ["qt_95_low", "qt_68_low", "median", "qt_68_high", "qt_95_high"]:
            axes[1].plot(
                results["mag_mean"], results[qt], "--", color="blue", linewidth=2.0
            )

        axes[1].set_xlabel("Magnitude")
        axes[1].set_ylabel(r"$(z_{phot} - z_{spec})/(1+z_{spec})$")

        plot_name = self._make_full_plot_name(prefix, "")

        return RailPlotHolder(
            name=plot_name, figure=figure, plotter=self, dataset_holder=dataset_holder
        )

    def _make_plots(self, prefix: str, **kwargs: Any) -> dict[str, RailPlotHolder]:
        find_only = kwargs.get("find_only", False)
        figtype = kwargs.get("figtype", "png")
        dataset_holder = kwargs.get("dataset_holder")
        out_dict: dict[str, RailPlotHolder] = {}
        truth: np.ndarray = kwargs["truth"]
        pointEstimate: np.ndarray = kwargs["pointEstimate"]
        magnitude: np.ndarray = kwargs["magnitude"]
        if find_only:
            plot_name = self._make_full_plot_name(prefix, "")
            assert dataset_holder
            plot = RailPlotHolder(
                name=plot_name,
                path=os.path.join(dataset_holder.config.name, f"{plot_name}.{figtype}"),
                plotter=self,
                dataset_holder=dataset_holder,
            )
        else:
            plot = self._make_biweight_stats_plot(
                prefix=prefix,
                truth=truth,
                pointEstimate=pointEstimate,
                magnitude=magnitude,
                dataset_holder=dataset_holder,
            )
        out_dict[plot.name] = plot
        return out_dict

    def process_data(  # pylint: disable=too-many-arguments
        self,
        zphot: np.ndarray,
        specz: np.ndarray,
        mag: np.ndarray,
        low: float = 0.01,
        high: float = 2.0,
        nclip: int = 3,
        nbin: int = 101,
    ) -> dict[str, list[float] | np.ndarray]:
        dz = (zphot - specz) / (1 + specz)

        mag_bins = np.linspace(low, high, nbin)
        # Bin the data
        bin_indices = np.digitize(mag, bins=mag_bins) - 1  # Assign each point to a bin

        biweight_mean: list[float] = []
        biweight_std: list[float] = []
        biweight_sigma: list[float] = []
        biweight_outlier: list[float] = []

        qt_95_low: list[float] = []
        qt_68_low: list[float] = []
        median: list[float] = []
        qt_68_high: list[float] = []
        qt_95_high: list[float] = []
        for i in range(len(mag_bins) - 1):
            subset = dz[bin_indices == i]
            subset_clip, _, _ = sigmaclip(subset, low=3, high=3)
            for _j in range(nclip):
                subset_clip, _, _ = sigmaclip(subset_clip, low=3, high=3)

            if len(subset_clip) == 0:  # pragma: no cover
                biweight_mean.append(np.nan)
                biweight_std.append(np.nan)
                biweight_sigma.append(np.nan)
                biweight_outlier.append(np.nan)
                qt_95_low.append(np.nan)
                qt_68_low.append(np.nan)
                median.append(np.nan)
                qt_68_high.append(np.nan)
                qt_95_high.append(np.nan)
                continue

            biweight_mean.append(biweight_location(subset_clip))
            biweight_std.append(biweight_scale(subset_clip) / np.sqrt(len(subset_clip)))
            biweight_sigma.append(biweight_scale(subset_clip))

            outlier_rate = np.sum(
                np.abs(subset) > 3 * biweight_scale(subset_clip)
            ) / len(subset)
            biweight_outlier.append(outlier_rate)

            qt_95_low.append(np.percentile(subset, 2.5))
            qt_68_low.append(np.percentile(subset, 16))
            median.append(np.percentile(subset, 50))
            qt_68_high.append(np.percentile(subset, 84))
            qt_95_high.append(np.percentile(subset, 97.5))

        mag_mean = (mag_bins[:-1] + mag_bins[1:]) / 2

        return {
            "mag_mean": mag_mean,
            "biweight_mean": biweight_mean,
            "biweight_std": biweight_std,
            "biweight_sigma": biweight_sigma,
            "biweight_outlier": biweight_outlier,
            "qt_95_low": qt_95_low,
            "qt_68_low": qt_68_low,
            "median": median,
            "qt_68_high": qt_68_high,
            "qt_95_high": qt_95_high,
        }



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

from ceci.config import StageParameter
from rail.core.configurable import Configurable

from rail.projects.dynamic_class import DynamicClass

from .dataset import RailDataset
from .dataset_holder import RailDatasetHolder
from .plot_holder import RailPlotDict

if TYPE_CHECKING:
    from .plot_holder import RailPlotHolder
    from .plotter_factory import RailPlotterFactory


class RailPlotter(Configurable, DynamicClass):
    """Base class for making matplotlib plot

    The main function in this class is:

    .. highlight:: python
    .. code-block:: python

      run(prefix: str, kwargs**: Any) -> dict[str, RailPlotHolder]

    This function will make a set of plots and return them in a dict.
    prefix is string that gets prepended to plot names.

    The data to be plotted is passed in via the kwargs.


    Sub-classes should implement

    .. highlight:: python
    .. code-block:: python

      config_options: dict[str, ceci.StageParameter]

    that will be used to configure things like the axes binning, selection functions,
    and other plot-specfic options

    .. highlight:: python
    .. code-block:: python

      input_type: RailPZPointEstimateDataset

    that specifics the inputs
    that the sub-classes expect, this is used the check the kwargs
    that are passed to the `run` function.

    A function:

    .. highlight:: python
    .. code-block:: python

      _make_plots(self, prefix: str, **kwargs: Any) -> dict[str, RailPlotHolder]:

    That actually makes the plots.  It does not need to do the checking
    that the correct kwargs have been given.
    """

    input_type: type[RailDataset] = RailDataset

    sub_classes: dict[str, type[DynamicClass]] = {}

    yaml_tag = "Plotter"

    @staticmethod
    def iterate_plotters(
        name: str,
        plotters: list[RailPlotter],
        prefix: str,
        dataset: RailDatasetHolder,
        **kwargs: Any,
    ) -> RailPlotDict:
        """Utility function to run several plotters on the same data

        Parameters
        ----------
        name: str
            Name to give to the RailPlotDict

        plotters: list[RailPlotter]
            Plotters to run

        prefix: str
            Prefix to append to plot names, e.g., the p(z) algorithm or
            analysis 'flavor'

        kwargs: dict[str, Any]
            Used to pass the data to make the plots

        Returns
        -------
        out_dict: RailPlotDict
            Dictionary of the newly created figures
        """
        out_dict: dict[str, RailPlotHolder] = {}
        extra_args: dict[str, Any] = dict(dataset_holder=dataset)
        for plotter_ in plotters:
            out_dict.update(
                plotter_.run(prefix, **dataset.resolve(), **kwargs, **extra_args)
            )
        return RailPlotDict(name=name, plots=out_dict)

    @staticmethod
    def iterate(
        plotters: list[RailPlotter],
        datasets: list[RailDatasetHolder],
        **kwargs: Any,
    ) -> dict[str, RailPlotDict]:
        """Utility function to several plotters of several data sets

        Parameters
        ----------
        plotters: list[RailPlotter]
            Plotters to run

        datasets: list[RailDatasetHolder]
            Prefixes and datasets to iterate over

        Returns
        -------
        out_dict: dict[str, RailPlotDict]
            Dictionary of the newly created figures
        """
        out_dict: dict[str, RailPlotDict] = {}
        for val in datasets:
            out_dict[val.config.name] = RailPlotter.iterate_plotters(
                val.config.name, plotters, "", val, **kwargs
            )
        return out_dict

    @staticmethod
    def write_plots(
        fig_dict: dict[str, RailPlotDict],
        outdir: str = ".",
        figtype: str = "png",
        purge: bool = False,
    ) -> None:
        """Utility function to write several plots do disk

        Parameters
        ----------
        fig_dict: dict[str, RailPlotDict]
            Dictionary of figures to write

        outdir: str
            Directory to write figures in

        figtype: str
            Type of figures to write, e.g., png, pdf...

        purge: bool
            Delete figure after saving
        """
        for key, val in fig_dict.items():
            try:
                os.makedirs(outdir)
            except Exception:
                pass
            out_path = os.path.join(outdir, key)
            val.savefigs(out_path, figtype=figtype, purge=purge)

    def __init__(self, **kwargs: Any):
        """C'tor

        Parameters
        ----------
        kwargs: Any
            Configuration parameters for this plotter, must match
            class.config_options data members
        """
        DynamicClass.__init__(self)
        Configurable.__init__(self, **kwargs)

    def __repr__(self) -> str:
        return f"{type(self)}"

    def run(
        self,
        prefix: str,
        **kwargs: dict[str, Any],
    ) -> dict[str, RailPlotHolder]:
        """Make all the plots given the data

        Parameters
        ----------
        prefix: str
            Prefix to append to plot names, e.g., the p(z) algorithm or
            analysis 'flavor'

        kwargs: dict[str, Any]
            Used to pass the data to make the plots

        Returns
        -------
        out_dict: dict[str, RailPlotHolder]
            Dictionary of the newly created figures
        """
        self._validate_inputs(**kwargs)
        return self._make_plots(prefix, **kwargs)

    def _make_full_plot_name(self, prefix: str, plot_name: str) -> str:
        """Create the make for a specific plot

        Parameters
        ----------
        prefix: str
            Prefix to append to plot names, e.g., the p(z) algorithm or
            analysis 'flavor'

        plot_name: str
            Specific name for a particular plot

        Returns
        -------
        plot_name: str
            Plot name, following the pattern f"{prefix}{self._name}{plot_name}"
        """
        return f"{prefix}{self.config.name}{plot_name}"

    def to_yaml_dict(self) -> dict[str, dict[str, Any]]:
        """Create a yaml-convertable dict for this object"""
        yaml_dict = Configurable.to_yaml_dict(self)
        yaml_dict[self.yaml_tag].update(class_name=f"{self.full_class_name()}")
        return yaml_dict

    @classmethod
    def _validate_inputs(cls, **kwargs: Any) -> None:
        cls.input_type.validate_inputs(**kwargs)

    def _make_plots(
        self,
        prefix: str,
        **kwargs: Any,
    ) -> dict[str, RailPlotHolder]:
        raise NotImplementedError()


class RailPlotterList(Configurable):
    """The class collects a set of plotter that can all run on the same data.

    E.g., plotters that can all run on a dict that looks like
    `{truth:np.ndarray, pointEstimates: np.ndarray}` could be put into a PlotterList.
    This make it easier to collect similar types of plots.
    """

    config_options: dict[str, StageParameter] = dict(
        name=StageParameter(str, None, fmt="%s", required=True, msg="PlotterList name"),
        dataset_holder_class=StageParameter(
            str,
            None,
            fmt="%s",
            required=True,
            msg="Dataset holder that provides datset types expected by plotter on the list",
        ),
        plotters=StageParameter(
            list,
            [],
            fmt="%s",
            msg="List of plotter to include",
        ),
    )

    yaml_tag = "PlotterList"

    def __init__(self, **kwargs: Any):
        """C'tor

        Parameters
        ----------
        kwargs: Any
            Configuration parameters for this RailPlotterListHolder, must match
            class.config_options data members
        """
        Configurable.__init__(self, **kwargs)

    def __repr__(self) -> str:
        return f"{self.config.plotters}"

    def resolve(self, plotter_factory: RailPlotterFactory) -> list[RailPlotter]:
        """Extract the plotters

        Paramters
        ---------
        plotter_factory:
            Factory used to make the plotters.

        Returns
        -------
        list[RailPlotter]
            Requested plotters.

        Notes
        -----
        This will enforce that each plotter expects the compatible dataset_types
        """

        the_list: list[RailPlotter] = []

        dataset_holder_class = RailDatasetHolder.load_sub_class(
            self.config.dataset_holder_class
        )
        dataset_class = dataset_holder_class.output_type

        for name_ in self.config.plotters:
            a_plotter = plotter_factory.get_plotter(name_)
            if not issubclass(dataset_class, a_plotter.input_type):  # pragma: no cover
                raise TypeError(
                    f"PlotterList dataset_class {dataset_class} is "
                    f"not a subclass of Plotter.input_type {a_plotter.input_type}."
                )
            the_list.append(a_plotter)
        return the_list
